import logging
import os
from typing import List

from caffoa import duplication_handler
from caffoa.base_writer import BaseWriter
from caffoa.body_type_filter import BodyTypeFilter
from caffoa.converter import get_response_type, is_primitive
from caffoa.model import EndPoint, Response, MethodResult


class FunctionWriter(BaseWriter):
    def __init__(self, version: int, name: str, namespace: str, target_folder: str, interface_name: str):
        super().__init__(version)
        self.interface_name = interface_name
        self.version = version
        self.target_folder = target_folder
        self.namespace = namespace
        self.boilerplate = "{BASE}"
        self.use_factory = True if version >= 3 else False
        self.json_error_handling = {"class": "CaffoaJsonParseError"}
        self.functions_name = f"{name}Functions"
        self.error_folder = os.path.join(target_folder, "Errors")
        self.error_namespace = namespace + ".Errors"
        self.request_body_filter = list()
        self.imports = list()
        self.route_prefix = ''
        self.method_template = self.load_template("FunctionMethod.cs")
        self.class_template = self.load_template("FunctionTemplate.cs")
        self.caffoa_error_template = self.load_template("CaffoaClientError.cs")
        self.client_error_template = self.load_template("ClientErrorTemplate.cs")
        self.generic_client_error_template = self.load_template("GenericClientErrorTemplate.cs")

    def write(self, endpoints: List[EndPoint]):
        os.makedirs(self.target_folder, exist_ok=True)
        imports = []
        imports.extend(self.imports)
        if self.version > 2:
            self.write_errors(endpoints)
            imports.append(self.error_namespace)
            json_error_namespace = self.json_error_handling.get("namespace", self.error_namespace)
            if json_error_namespace not in imports:
                imports.append(json_error_namespace)

        json_error_class = self.json_error_handling["class"]
        function_endpoints = []
        for ep in endpoints:
            function_endpoints.append(self.format_endpoint(ep))
#            imports.extend(BodyTypeFilter(self.request_body_filter).additional_imports(ep))
        interface_name = self.interface_name + "Factory" if self.version > 1 and self.use_factory else self.interface_name
        imports = [f"using {x};\n" for x in imports]
        file_name = os.path.abspath(f"{self.target_folder}/{self.functions_name}.generated.cs")
        logging.info(f"Writing Functions to {file_name}")
        with open(file_name, "w", encoding="utf-8") as f:
            f.write(self.class_template.format(METHODS="\n\n".join(function_endpoints),
                                               NAMESPACE=self.namespace,
                                               CLASSNAME=self.functions_name,
                                               INTERFACENAME=interface_name,
                                               IMPORTS="".join(imports),
                                               JSON_ERROR_CLASS=json_error_class))

    def format_endpoint(self, endpoint: EndPoint) -> str:
        if self.version == 1:
            template_params = self.v1_params(endpoint)
        if self.version == 2:
            template_params = self.v2_params(endpoint)
        if self.version == 3:
            template_params = self.v3_params(endpoint)
        template_params["PARAMS"] = ", ".join(template_params["PARAMS"])
        if len(template_params["PARAM_NAMES"]) > 0:
            template_params["PARAM_NAMES"] = ", " + (", ".join(template_params["PARAM_NAMES"]))
        else:
            template_params["PARAM_NAMES"] = ""
        template_params["CALL"] = template_params["CALL"].format_map(template_params)
        return self.method_template.format_map(template_params)

    def v1_params(self, endpoint: EndPoint) -> dict:
        template_params = self.default_params(endpoint)
        if endpoint.needs_content:
            template_params['PARAMS'].append("req.Content")
        params = ", ".join(template_params["PARAMS"])
        function_call = f"{endpoint.name}({params})"
        default_call = f"return await Service(req, log).{function_call};"
        template_params["INVOCATION"] = self.boilerplate.replace("{BASE}", default_call) \
            .replace("{CALL}", function_call) \
            .replace("\n", "\n\t\t\t\t").strip()
        return template_params

    def v2_params(self, endpoint: EndPoint) -> dict:
        template_params = self.default_params(endpoint)
        if self.use_factory:
            func_invocation = ".Instance(request)"
            if endpoint.needs_content:
                template_params['PARAMS'].append("request.Body")
        else:
            func_invocation = ""
            template_params['PARAMS'].append("request")
        params = ", ".join(template_params["PARAMS"])
        function_call = f"{endpoint.name}({params})"
        default_call = f"return await _service{func_invocation}.{function_call};"
        template_params["INVOCATION"] = self.boilerplate.replace("{BASE}", default_call) \
            .replace("{CALL}", function_call) \
            .replace("\n", "\n\t\t\t\t").strip()
        return template_params

    def v3_params(self, endpoint: EndPoint) -> dict:
        params = self.default_params(endpoint)
        type = get_response_type(endpoint)
        params['VALUE'] = f"var result = "
        filtered_body = BodyTypeFilter(self.request_body_filter).body_type(endpoint)

        if type is None:
            params['RESULT'] = "result"
        elif type.name is None and type.is_simple:
            params['RESULT'] = f"new StatusCodeResult({type.code})"
            params['VALUE'] = ""
        elif type.name is None:
            params['RESULT'] = f"new StatusCodeResult(result)"
        elif type.is_simple:
            result = self.get_result_handler(type)
            params['RESULT'] = f"new JsonResult({result}) {{StatusCode = {type.code}, SerializerSettings = _service.ResponseSerializerSettings }}"
        elif type.base is None:
            params['RESULT'] = "new StatusCodeResult(code)"
            params['VALUE'] = "var code = "
        else:
            params['RESULT'] = f"new JsonResult(result) {{StatusCode = code, SerializerSettings = _service.ResponseSerializerSettings  }}"
            params['VALUE'] = f"var (result, code) = "
        error_infos = [f'("{param.name}", {param.name})' for param in
                       endpoint.parameters]
        if len(error_infos) > 0:
            params['ADDITIONAL_ERROR_INFOS'] = ", " + (", ".join(error_infos))
        if self.use_factory:
            params["FACTORY_CALL"] = "Instance(request)."
        if endpoint.needs_content and endpoint.body is not None and endpoint.body.is_selection():
            params = self.v3_params_selection(endpoint, params)
        elif endpoint.needs_content and filtered_body is not None:
            params['PARAMS'].append(f"await ParseJson<{filtered_body}>(request.Body)")
        elif endpoint.needs_content and endpoint.body is not None:
            params['PARAMS'].append(f"await ParseJson<{endpoint.body.types[0]}>(request.Body)")
        elif endpoint.needs_content and self.use_factory:
            params['PARAMS'].append("request.Body")
        if not self.use_factory:
            params['PARAMS'].append("request")

        return params

    def v3_params_selection(self, endpoint: EndPoint, template_params: dict) -> dict:
        params = list(template_params["PARAMS"]).copy()
        call_params = template_params.copy()
        call_params["VALUE"] = ""
        cases = list()
        allowed_values = list()
        for value,typename in endpoint.body.mapping.items():
            option_params = params.copy()
            option_params.append(f"ToObject<{typename}>(jObject)")
            if not self.use_factory:
                option_params.append(f"request")
            call_params["PARAMS"] = ", ".join(option_params)
            call = "await _service.{FACTORY_CALL}{NAME}({PARAMS})".format_map(call_params)
            cases.append(f'"{value}" => {call}')
            allowed_values.append(f'"{value}"')
        template = self.load_template("FunctionSwitchTemplate.cs")
        template_params["CASES"] = ",\n\t\t\t\t\t".join(cases)
        template_params["CASES_ALLOWED_VALUES"] = ", ".join(allowed_values)
        template_params["DISC"] = endpoint.body.discriminator
        template_params["JSON_ERROR_CLASS"] = self.json_error_handling.get("class")
        template_params["CALL"] = template.format_map(template_params)
        return template_params




    def default_params(self, endpoint: EndPoint) -> dict:
        extra_error_info = [f'debugInformation["p_{param.name}"] = {param.name}.ToString();\n\t\t\t\t' for param in
                            endpoint.parameters]
        return dict(
            NAME=endpoint.name,
            OPERATION=endpoint.method,
            PATH=self.route_prefix + endpoint.path,
            PARAMS=[param.name for param in endpoint.parameters],
            PARAM_NAMES=[f"{param.type} {param.name}" for param in endpoint.parameters],
            ADDITIONAL_ERROR_INFOS="".join(extra_error_info),
            INVOCATION="",
            RESULT="",
            VALUE="",
            FACTORY_CALL="",
            CALL="{VALUE}await _service.{FACTORY_CALL}{NAME}({PARAMS});"
        )

    def write_errors(self, endpoints: List[EndPoint]):
        file_name = os.path.abspath(f"{self.error_folder}/CaffoaClientError.generated.cs")
        logging.info(f"Writing Client Error to {file_name}")
        with open(file_name, "w", encoding="utf-8") as f:
            f.write(self.caffoa_error_template.format(NAMESPACE=self.error_namespace))
        return
        error_classes = dict()
        generic_error_classes = dict()
        for endpoint in endpoints:
            if endpoint.responses is None:
                continue
            for response in endpoint.responses:
                if 400 <= response.code < 500:
                    name = response.content
                    if name is None:
                        generic_error_classes[f'Generic{response.code}'] = response.code
                        continue
                    if name in error_classes:
                        if error_classes[name] != response.code:
                            raise Warning(
                                "The same response object with different error codes is currently not supported")
                    error_classes[name] = response.code
        # remove duplicates but keep order:
        imports = list(dict.fromkeys(imports))
        imports_str = "".join([f"using {imp};\n" for imp in self.imports])
        os.makedirs(self.error_folder, exist_ok=True)


        for name, code in error_classes.items():
            if duplication_handler.should_generate("Error/" + name):
                duplication_handler.store_generated("Error/" + name)
                file_name = os.path.abspath(f"{self.error_folder}/{name}ClientError.generated.cs")
                logging.info(f"Writing Client Error to {file_name}")
                with open(file_name, "w", encoding="utf-8") as f:
                    f.write(self.client_error_template.format(NAMESPACE=self.error_namespace,
                                                              NAME=name,
                                                              CODE=code,
                                                              IMPORTS=imports_str))
        for name, code in generic_error_classes.items():
            if duplication_handler.should_generate("Error/" + name):
                duplication_handler.store_generated("Error/" + name)
                file_name = os.path.abspath(f"{self.error_folder}/{name}ClientError.generated.cs")
                logging.info(f"Writing Client Error to {file_name}")
                with open(file_name, "w", encoding="utf-8") as f:
                    f.write(self.generic_client_error_template.format(NAMESPACE=self.error_namespace,
                                                                      NAME=name,
                                                                      CODE=code))

    def get_result_handler(self, result: MethodResult):
        return "result"
        base = result.base
        is_array = False
        if base.startswith("IEnumerable<"):
            base = base[12:-1]
            is_array = True
        if is_primitive(base):
            return "result"
        if is_array:
            return f"result.Select(o=>o.To{base}())"
        return f"result.To{result.name}()"





import argparse
import glob
import logging
import os

import yaml

from caffoa import duplication_handler
from caffoa.openapi_file import OpenApiFile


def execute():
    parser = argparse.ArgumentParser()
    parser.add_argument("--config", default="caffoa.yml", help="Path to config file (Default: caffoa.yml)")
    parser.add_argument('--clean', action='store_true', help="remove all files with the pattern *.generated.cs")
    args = parser.parse_args()
    with open(args.config, "r", encoding="utf-8") as f:
        data = yaml.safe_load(f)
    if not "services" in data:
        raise Warning("no services in config file.")
    services = data["services"]
    if type(services) != list:
        raise Warning("services should be list")
    try:
        settings = data["config"]
        if type(settings) != dict:
            raise Warning("config should be key/values pairs")
    except KeyError:
        settings = dict()
    version = settings.get("version", 1)
    duplication_handler.init(settings.get("duplicates", "overwrite"))

    if settings.get('clearGeneratedFiles', args.clean):
        old_files = glob.glob('**/*.generated.cs', recursive=True)
        for file in old_files:
            logging.info(f"Removing {file}")
            os.unlink(file)


    for number, config in enumerate(services):
        if not "apiPath" in config:
            raise Warning(f"apiPath is required for service #{number}")
        api_config = settings.copy()
        api_config.update(config.get("config", dict()))
        handler = OpenApiFile(config["apiPath"], version, api_config)
        if "model" in config:
            handler.create_model(config["model"])

        if "function" in config:
            handler.create_function(config["function"])



import logging
import re
from typing import Optional, Tuple

from caffoa.converter import to_camelcase, parse_type, is_date
from caffoa.model import ModelData, MemberData, ModelObjectData, ModelInterfaceData


class BaseObjectParser:
    def __init__(self, prefix: str, suffix: str):
        self.suffix = suffix
        self.prefix = prefix

    def name_from_ref(self, param: str):
        if ".yml_schemas_" in param:
            param = param.split(".yml_schemas_")[-1]
        name = param.split('/')[-1]
        return name

    def class_name_from_ref(self, param: str):
        return self.class_name(self.name_from_ref(param))

    def class_name(self, name):
        return self.prefix + to_camelcase(name) + self.suffix


class ObjectParser(BaseObjectParser):
    def __init__(self, raw_name, prefix: str, suffix: str, known_types: dict):
        super().__init__(prefix, suffix)
        self.name = raw_name
        self.known_types = known_types
        self.result = ModelObjectData(self.name, self.class_name(self.name))

    def parse(self, schema: dict) -> ModelData:
        if "allOf" in schema:
            schema, self.result.parent = self.handle_all_of(schema)
        if "oneOf" in schema:
            self.result = ModelInterfaceData(self.name, self.class_name(self.name))
            discriminator = schema.get("discriminator")
            if discriminator:
                self.result.discriminator = to_camelcase(discriminator["propertyName"])
            for element in schema["oneOf"]:
                if "$ref" in element:
                    self.result.children.append(self.class_name_from_ref(element["$ref"]))
                elif "required" in element:
                    pass # this is fine: https://github.com/p1c2u/openapi-spec-validator/issues/27
                else:
                    logging.warning("did not find $ref as child of oneOf")
        elif "properties" in schema:
            required_props = schema.get('required', list())
            for name, data in schema["properties"].items():
                self.result.properties.append(self.create_param(name, data, name in required_props))

        if "description" in schema:
            self.result.description = schema["description"]

        return self.result

    def handle_all_of(self, schema: dict) -> Tuple[dict, Optional[str]]:
        to_generate = None
        parent = None
        for element in schema["allOf"]:
            if "$ref" in element:
                if parent is None:
                    parent = self.class_name_from_ref(element["$ref"])
                else:
                    raise Warning("allOf is implemented as inheritance; cannot have to children with $ref")
            if "type" in element or "properties" in element:
                if to_generate is not None:
                    raise Warning(
                        "allOf is implemented as inheritance; cannot have to children of allOf with direct implementation")
                to_generate = element
        if to_generate is None:
            logging.warning("Creating class without content, no child of allOf is type object")
            to_generate = dict()
        return to_generate, parent

    def create_param(self, name: str, data: dict, required: bool) -> MemberData:
        param = MemberData(name)
        # handle default type references, replace data if so
        if "$ref" in data:
            param.typename = self.class_name_from_ref(data["$ref"])
            if param.typename in self.known_types:
                data = self.known_types[param.typename]

        param.is_required = required
        param.nullable = "nullable" in data and data["nullable"]

        # handle nullable references that are constructed with anyOf
        if "anyOf" in data:
            data = self.handle_any_of(data, name)

        if "description" in data:
            param.description = data["description"]

        if "$ref" in data:
            param.typename, param.default_value = self.handle_ref(data, param.is_required)
            param.nullable = not param.is_required
            param.is_basic_type = False

        elif "type" not in data:
            raise Warning(f"Cannot parse property without type for '{name}' in '{self.name}'")

        elif data["type"] == "array":
            param.typename, param.default_value = self.handle_array(data, name, param.nullable)
            self.result.imports.append("System.Collections.Generic")
            self.result.imports.append("System.Linq")
            param.is_list = True

        elif data["type"] == "object":
            if "properties" in data:
                raise Warning(
                    f"Cannot parse property trees: '{name}' child of '{self.name}' should be declared in own schema directly under 'components'")
            if "additionalProperties" in data:
                param.typename, param.default_value = self.handle_additional_properties(data, param.nullable)
                self.result.imports.append("System.Collections.Generic")
            else:
                raise Warning(
                    f"Cannot nested object without additionalProperties: '{name}' child of '{self.name}'")

        else:
            param.typename, param.default_value = self.handle_default_type(data, param.nullable)
            param.enums = self.handle_enums(name, data)
            if param.enums and None in param.enums.keys():
                param.typename, param.default_value = self.handle_default_type(data, True)
            if param.enums is not None:
                self.result.imports.append("System")
                self.result.imports.append("System.Collections.Immutable")
                self.result.imports.append("System.Linq")
            param.is_date = is_date(data)
        return param

    def handle_any_of(self, data: dict, name: str):
        anyof = data["anyOf"]
        if len(anyof) != 1:
            raise Warning(f"anyOf with multiple children not supported for '{name}' in '{self.name}'")
        anyof = anyof[0]
        if "$ref" not in anyof:
            raise Warning(f"anyOf child other than $ref not supported for '{name}' in '{self.name}'")
        data["$ref"] = anyof["$ref"]
        typename = self.class_name_from_ref(data["$ref"])
        if typename in self.known_types:
            data = self.known_types[typename]
        return data

    def handle_ref(self, data: dict, required: bool) -> Tuple[str, str]:
        typename = self.class_name_from_ref(data["$ref"])
        if required:
            default_value = f"new {typename}()"
        else:
            default_value = "null"
        return typename, default_value

    def handle_array(self, data: dict, name: str, nullable: bool) -> Tuple[str, str]:
        items = data["items"]
        if "$ref" in items:
            typename = self.class_name_from_ref(items["$ref"])
        elif items["type"] not in ["string", "number", "integer", "boolean"]:
            raise Warning(
                f"Cannot parse array trees: '{name}' child of '{self.name}' should have it's array item declared in own schema directly under 'components'")
        else:
            typename = parse_type(items, nullable)
        return f"ICollection<{typename}>", f"new List<{typename}>()"

    def handle_additional_properties(self, data: dict, nullable: bool) -> Tuple[str, str]:
        props = data['additionalProperties']
        if "$ref" in props:
            typename = self.class_name_from_ref(props["$ref"])
        else:
            typename = parse_type(props, nullable)
        return f"Dictionary<string, {typename}>", f"new Dictionary<string, {typename}>()"

    def handle_default_type(self, data: dict, nullable: bool) -> Tuple[str, str]:
        default_value = None
        if "default" in data:
            default_value = data["default"]
            if type(default_value) is str:
                default_value = f'"{default_value}"'
            elif type(default_value) is bool:
                default_value = "true" if default_value else "false"
            elif default_value is None:
                default_value = "null"
        elif nullable:
            default_value = "null"
        typename = parse_type(data, nullable and default_value == "null")
        return typename, default_value

    def handle_enums(self, name: str, data: dict) -> Optional[dict]:
        if "enum" not in data:
            return None
        enum = data["enum"]
        if len(enum) == 0:
            return None
        typename, _ = self.handle_default_type(data, False)
        if typename not in ["string", "int", "long", "double"]:
            return None

        enums = dict()
        for value in enum:
            name = value
            name = re.sub(r"[^A-Za-z0-9]+", '_', str(name))
            if typename == "string":
                value = f'"{value}"'
            enums[value] = name

        return enums



import os


class BaseWriter:
    def __init__(self, version: int):
        self.version = version
        self.template_folder = os.path.join(os.path.dirname(os.path.abspath(__file__)), f"data/v{version}")
        self.base_template_folder = os.path.join(os.path.dirname(os.path.abspath(__file__)), f"data/base")

    def load_template(self, name):
        path = os.path.join(self.template_folder, name)
        if not os.path.exists(path):
            path = os.path.join(self.base_template_folder, name)

        with open(path, "r", encoding="utf-8") as f:
            return f.read()



import logging
import os
from typing import List

from caffoa.base_writer import BaseWriter
from caffoa.body_type_filter import BodyTypeFilter
from caffoa.converter import get_response_type
from caffoa.model import EndPoint


class InterfaceWriter(BaseWriter):
    def __init__(self, version: int, name: str, namespace: str, target_folder: str):
        super().__init__(version)
        self.version = version
        self.target_folder = target_folder
        self.use_factory = True if version >= 3 else False
        self.namespace = namespace
        self.imports = list()
        self.request_body_filter = list()
        self.interface_name = f"I{name}Service"
        self.interface_method_template = self.load_template("InterfaceMethod.cs")
        self.interface_template = self.load_template("InterfaceTemplate.cs")
        self.factory_interface_template = self.load_template("FactoryInterfaceTemplate.cs")

    def write(self, endpoints: List[EndPoint]):
        os.makedirs(self.target_folder, exist_ok=True)

        interfaces = []
        imports = ["System.Collections.Generic"]
        for ep in endpoints:
            interfaces.extend(self.format_endpoints(ep))
            imports.extend(BodyTypeFilter(self.request_body_filter).additional_imports(ep))
        imports.extend(self.imports)
        # remove duplicates but keep order:
        imports = list(dict.fromkeys(imports))
        imports_str = "".join([f"using {imp};\n" for imp in imports])

        file_name = os.path.abspath(f"{self.target_folder}/{self.interface_name}.generated.cs")
        logging.info(f"Writing Interface to {file_name}")
        with open(file_name, "w", encoding="utf-8") as f:
            f.write(self.interface_template.format(METHODS="\n\n".join(interfaces),
                                                   NAMESPACE=self.namespace,
                                                   IMPORTS=imports_str,
                                                   CLASSNAME=self.interface_name))
        if self.version > 1 and self.use_factory:
            file_name = os.path.abspath(f"{self.target_folder}/{self.interface_name}Factory.generated.cs")
            logging.info(f"Writing Factory Interface to {file_name}")
            with open(file_name, "w", encoding="utf-8") as f:
                f.write(self.factory_interface_template.format(NAMESPACE=self.namespace,
                                                               CLASSNAME=self.interface_name))

    def format_endpoints(self, endpoint: EndPoint) -> List[str]:
        base_params = [f"{param.type} {param.name}" for param in endpoint.parameters]
        if self.version == 1:
            params = base_params.copy()
            if endpoint.needs_content:
                params.append("HttpContent contentPayload")
            return [self.format_endpoint(endpoint, params, "")]

        type = get_response_type(endpoint)
        if type is None:
            result = "<IActionResult>"
        elif type.name is None:
            result = ""
        else:
            result = f"<{type.name}>"
        params = base_params.copy()
        filtered_body = BodyTypeFilter(self.request_body_filter).body_type(endpoint)
        if filtered_body:
            params.append(f"{filtered_body} payload")
            if not self.use_factory:
                params.append("HttpRequest request")
            return [self.format_endpoint(endpoint, params, result)]
        if not endpoint.body:
            if endpoint.needs_content and self.use_factory:
                params.append(f"Stream stream")
            if not self.use_factory:
                params.append("HttpRequest request")
            return [self.format_endpoint(endpoint, params, result)]
        interfaces = list()
        for body in endpoint.body.types:
            params = base_params.copy()
            params.append(f"{body} payload")
            if not self.use_factory:
                params.append("HttpRequest request")
            interfaces.append(self.format_endpoint(endpoint, params, result))
        return interfaces

    def format_endpoint(self, endpoint: EndPoint, params: list, result: str):
        formatted_params = ", ".join(params)
        return self.interface_method_template.format(NAME=endpoint.name,
                                                     PARAMS=formatted_params,
                                                     RESULT=result,
                                                     DOC="\n\t\t/// ".join(endpoint.documentation))
